/*jslint indent:2, maxlen: 80, nomen: true */
/*global jIO, RSVP, UriTemplate, SimpleQuery, ComplexQuery, QueryFactory,
  Query*/
(function (jIO, RSVP, UriTemplate, SimpleQuery, ComplexQuery, QueryFactory,
  Query) {
  "use strict";

  function inverseMappingDict(_mapping_dict, key) {
    var new_mapping_dict = {},
      new_key,
      new_value = {},
      new_kv = {};

    if (_mapping_dict[key].value !== undefined) {
      new_key = Object.keys(_mapping_dict[key].value)[0];
      new_value = inverseMappingDict(_mapping_dict[key].value,
                                                       new_key);
    }

    if (_mapping_dict[key].equal !== undefined) {
      if (typeof _mapping_dict[key].equal === "string") {
        new_mapping_dict[_mapping_dict[key].equal] = {'equal' : key};
        if (new_value !== undefined) {
          new_mapping_dict[_mapping_dict[key].equal].value = new_value;
        }
        return new_mapping_dict;
      }
      //xx: is it possible for equal to have children?
    }

    if (_mapping_dict[key].regex !== undefined) {
      new_mapping_dict = {
        regexp: _mapping_dict[key].regex,
        match_value: key
      };
      return new_mapping_dict;
    }

    if (_mapping_dict[key].multiEqual !== undefined) {
      new_mapping_dict = {
        multiEqualArray: _mapping_dict[key].multiEqual,
        match_value: key
      };
      return new_mapping_dict;
    }

    if (key === "switch") {
      //xx: assuming switch is just used in case of value
      new_mapping_dict['switch'] = {};
      new_mapping_dict['switch'].regexps = [];
      new_mapping_dict['switch'].multiEqual = [];
      for (new_key in _mapping_dict['switch']) {
        if (_mapping_dict['switch'].hasOwnProperty(new_key)) {
          new_kv = inverseMappingDict(_mapping_dict['switch'], new_key);
          new_key = Object.keys(new_kv)[0];
          if (new_key === "regexp" || new_key === "multiEqualArray") {
            new_mapping_dict['switch'].regexps.push(new_kv);
          } else {
            new_mapping_dict['switch'][new_key] = new_kv[new_key];
          }
        }
      }
      return new_mapping_dict;
    }
    new_mapping_dict[key] = {};
    new_mapping_dict[key].value = new_value;
    return new_mapping_dict;
  }

  function MappingStorage(spec) {
    this._mapping_dict = spec.mapping_dict || {};
    this._sub_storage = jIO.createJIO(spec.sub_storage);
    this._map_all_property = spec.map_all_property !== undefined ?
        spec.map_all_property : true;
    this._attachment_mapping_dict = spec.attachment_mapping_dict || {};
    this._query = spec.query || {};

    if (this._query.query !== undefined) {
      this._query.query = QueryFactory.create(this._query.query);
    }
    this._default_mapping = {};
    this._id_is_mapped = (this._mapping_dict.id !== undefined
            && this._mapping_dict.id.equal !== "id");
    var property, query_list = [], new_key,
      new_kv = {}, inverse_mapping_dict = {};

    // handle default_value.
    for (property in this._mapping_dict) {
      if (this._mapping_dict.hasOwnProperty(property)) {
        if (this._mapping_dict[property].default_value !== undefined) {
          this._default_mapping[property] =
            this._mapping_dict[property].default_value;
          query_list.push(new SimpleQuery({
            key: property,
            value: this._mapping_dict[property].default_value,
            type: "simple"
          }));
        }
        new_kv = inverseMappingDict(this._mapping_dict, property);
        new_key = Object.keys(new_kv)[0];
        inverse_mapping_dict[new_key] = new_kv[new_key];
      }
    }
    this._inverse_mapping_dict = inverse_mapping_dict;

    if (this._query.query !== undefined) {
      query_list.push(QueryFactory.create(this._query.query));
    }

    if (query_list.length > 1) {
      this._query.query = new ComplexQuery({
        type: "complex",
        query_list: query_list,
        operator: "AND"
      });
    } else if (query_list.length === 1) {
      this._query.query = query_list[0];
    }
  }

  function getAttachmentId(storage, sub_id, attachment_id, method) {
    var mapping_dict = storage._attachment_mapping_dict;
    if (mapping_dict !== undefined
        && mapping_dict[attachment_id] !== undefined
        && mapping_dict[attachment_id][method] !== undefined
        && mapping_dict[attachment_id][method].uri_template !== undefined) {
      return UriTemplate.parse(
        mapping_dict[attachment_id][method].uri_template
      ).expand({id: sub_id});
    }
    return attachment_id;
  }

  function getSubStorageId(storage, id, sub_doc) {
    var query;
    return new RSVP.Queue()
      .push(function () {
        if (!storage._id_is_mapped || id === undefined) {
          return id;
        }
        if (storage._mapping_dict.id.equal !== undefined) {
          if (JSON.stringify(storage._mapping_dict)
              .indexOf('{"equal":"id"}') >= 0
              && sub_doc !== undefined) {
            if (sub_doc.id !== undefined) {
              return sub_doc.id;
            }
            throw new jIO.util.jIOError(
              "Cannot find id field related",
              400
            );
          }
          query = new SimpleQuery({
            key: storage._mapping_dict.id.equal,
            value: id,
            type: "simple"
          });
          if (storage._query.query !== undefined) {
            query = new ComplexQuery({
              operator: "AND",
              query_list: [query, storage._query.query],
              type: "complex"
            });
          }
          query = Query.objectToSearchText(query);
          return storage._sub_storage.allDocs({
            "query": query,
            "sort_on": storage._query.sort_on,
            "select_list": storage._query.select_list,
            "limit": storage._query.limit
          })
            .push(function (data) {
              if (data.data.rows.length === 0) {
                throw new jIO.util.jIOError(
                  "Can not find id",
                  404
                );
              }
              if (data.data.rows.length > 1) {
                throw new TypeError("id must be unique field: " + id
                  + ", result:" + data.data.rows.toString());
              }
              return data.data.rows[0].id;
            });
        }
        throw new jIO.util.jIOError(
          "Unsuported option: " + storage._mapping_dict.id,
          400
        );
      });
  }

  function mapProperty(_mapping_dict, key, value, for_value, isQuery) {
    var new_prop,
      i,
      k,
      regexp,
      equalityArray,
      new_mapping_dict = {},
      mapped_kv = {};

    if (_mapping_dict !== undefined) {
      if (_mapping_dict.value !== undefined) {
        if (typeof _mapping_dict.value === "string") {
          new_prop = _mapping_dict.value;
          new_mapping_dict = _mapping_dict.value;
        } else {
          //XX: Find other usecases of switch and remove this block
          new_prop = Object.keys(_mapping_dict.value)[0];
          if (new_prop === "switch") {
            new_mapping_dict = _mapping_dict.value;
            key = value;
          } else {
            new_mapping_dict = _mapping_dict.value[new_prop];
          }
        }

        value = mapProperty(new_mapping_dict, key, value, true, isQuery).value;
      }

      if (_mapping_dict.equal !== undefined) {
        if (typeof _mapping_dict.equal === "string") {
          new_prop = _mapping_dict.equal;
          new_mapping_dict = _mapping_dict.equal;
        } else {
          new_prop = Object.keys(_mapping_dict.equal)[0];
          new_mapping_dict = _mapping_dict.equal[new_prop];
        }
        return mapProperty(new_mapping_dict, key, value, for_value, isQuery);
      }

      if (_mapping_dict.regex !== undefined || _mapping_dict.regexps) {
        if (_mapping_dict.regexps && _mapping_dict.regexps.length) {
          for (i = 0; i < _mapping_dict.regexps.length; i = i + 1) {
            try {
              regexp = new RegExp(_mapping_dict.regexps[i].regexp);
            } catch (e) {
              throw new jIO.util.jIOError(
                "Regex for key " + key + ":" + e,
                400
              );
            }
            if (regexp.test(value) === true) {
              mapped_kv.key = key;
              mapped_kv.value = _mapping_dict.regexps[i].match_value;
              return mapped_kv;
            }
          }
        }
        if (_mapping_dict.default_value === undefined) {
          throw new jIO.util.jIOError(
            "default_value should be defined for regex: " + _mapping_dict,
            400
          );
        }

        mapped_kv.key = key;
        mapped_kv.value = _mapping_dict.default_value;
        return mapped_kv;
      }

      if (_mapping_dict.multiEqual !== undefined) {
        if (_mapping_dict.multiEqual  && _mapping_dict.multiEqual.length) {
          for (i = 0; i < _mapping_dict.multiEqual.length; i = i + 1) {
            equalityArray = _mapping_dict.multiEqual[i].multiEqualArray;
            if (equalityArray === undefined) {
              break;
            }
            for (k = 0; k < equalityArray.length; k = k + 1) {
              if (value === equalityArray[k]) {
                mapped_kv.key = key;
                mapped_kv.value = _mapping_dict.multiEqual[i].match_value;
                return mapped_kv;
              }
            }
          }
        }
        mapped_kv.key = key;
        mapped_kv.value = _mapping_dict.default_value;
        if (isQuery === true) {
          mapped_kv.value = _mapping_dict.multiEqual;
        }
        return mapped_kv;
      }

      if (_mapping_dict["switch"] !== undefined) {
        if (_mapping_dict["switch"].hasOwnProperty(key)) {
          return mapProperty(_mapping_dict['switch'][key], key, value,
                            true, isQuery);
        }
        if (_mapping_dict["switch"].regexps) {
          new_mapping_dict.regexps = _mapping_dict["switch"].regexps;
          return mapProperty(new_mapping_dict, key, value, true, isQuery);
        }
        mapped_kv.key = key;
        mapped_kv.value = value;
        return mapped_kv;
      }

      if (typeof _mapping_dict === "string") {
        if (for_value === false) {
          mapped_kv.key = _mapping_dict;
          mapped_kv.value = value;
          return mapped_kv;
        }
        mapped_kv.key =  key;
        mapped_kv.value = _mapping_dict;
        return mapped_kv;
      }

      if (_mapping_dict.default_value !== undefined) {
        mapped_kv.key = key;
        mapped_kv.value = _mapping_dict.default_value;
        return mapped_kv;
      }

    }
    mapped_kv.key = key;
    mapped_kv.value = value;
    return mapped_kv;
  }

  function mapToMainDocument(storage, sub_doc, sub_id, delete_id_from_doc) {
    var doc = {},
      property,
      new_kv = {};

    if (sub_id) {
      sub_doc.id = sub_id;
    }

    if (storage._map_all_property) {
      for (property in sub_doc) {
        if (sub_doc.hasOwnProperty(property)) {
          new_kv = mapProperty(storage._inverse_mapping_dict[property],
                                     property, sub_doc[property], false);
          doc[new_kv.key] = new_kv.value;
        }
      }
    }
    if (delete_id_from_doc) {
      delete doc.id;
    }
    return doc;
  }

  function mapToSubstorageDocument(storage, doc, id, delete_id_from_doc) {
    var sub_doc = {},
      mapped_kv = {},
      property;

    doc.id = id;
    for (property in doc) {
      if (doc.hasOwnProperty(property)) {
        mapped_kv = mapProperty(storage._mapping_dict[property],
                                     property, doc[property], false);
        if (!mapped_kv && storage._map_all_property) {
          sub_doc[property] = doc[property];
        } else {
          sub_doc[mapped_kv.key] = mapped_kv.value;
        }
      }
    }
    for (property in storage._default_mapping) {
      if (storage._default_mapping.hasOwnProperty(property)) {
        sub_doc[property] = storage._default_mapping[property];
      }
    }
    if (delete_id_from_doc) {
      delete sub_doc.id;
    }
    return sub_doc;
  }

  function handleAttachment(context, argument_list, method) {
    return getSubStorageId(context, argument_list[0])
      .push(function (sub_id) {
        argument_list[0] = sub_id;
        argument_list[1] = getAttachmentId(
          context,
          sub_id,
          argument_list[1],
          method
        );
        return context._sub_storage[method + "Attachment"].apply(
          context._sub_storage,
          argument_list
        );
      });
  }

  MappingStorage.prototype.get = function (id) {
    var context = this;
    return getSubStorageId(this, id)
      .push(function (sub_id) {
        return context._sub_storage.get(sub_id)
          .push(function (sub_doc) {
            return mapToMainDocument(context, sub_doc, sub_id, true);
          });
      });
  };

  MappingStorage.prototype.post = function (doc) {
    if (!this._id_is_mapped) {
      return this._sub_storage.post(mapToSubstorageDocument(
        this,
        doc,
        true,
        true
      ));
    }
    throw new jIO.util.jIOError(
      "post is not supported with id mapped",
      400
    );
  };

  MappingStorage.prototype.put = function (id, doc) {
    var context = this,
      sub_doc = mapToSubstorageDocument(this, doc, id, false);
    return getSubStorageId(this, id, sub_doc)
      .push(function (sub_id) {
        delete sub_doc.id;
        return context._sub_storage.put(sub_id, sub_doc);
      })
      .push(undefined, function (error) {
        if (error instanceof jIO.util.jIOError && error.status_code === 404) {
          delete sub_doc.id;
          return context._sub_storage.post(sub_doc);
        }
        throw error;
      })
      .push(function () {
        return id;
      });
  };

  MappingStorage.prototype.remove = function (id) {
    var context = this;
    return getSubStorageId(this, id)
      .push(function (sub_id) {
        return context._sub_storage.remove(sub_id);
      })
      .push(function () {
        return id;
      });
  };

  MappingStorage.prototype.putAttachment = function (id, attachment_id) {
    return handleAttachment(this, arguments, "put", id)
      .push(function () {
        return attachment_id;
      });
  };

  MappingStorage.prototype.getAttachment = function () {
    return handleAttachment(this, arguments, "get");
  };

  MappingStorage.prototype.removeAttachment = function (id, attachment_id) {
    return handleAttachment(this, arguments, "remove", id)
      .push(function () {
        return attachment_id;
      });
  };

  MappingStorage.prototype.allAttachments = function (id) {
    var context = this, sub_id;
    return getSubStorageId(context, id)
      .push(function (sub_id_result) {
        sub_id = sub_id_result;
        return context._sub_storage.allAttachments(sub_id);
      })
      .push(function (result) {
        var attachment_id,
          attachments = {},
          mapping_dict = {};
        for (attachment_id in context._attachment_mapping_dict) {
          if (context._attachment_mapping_dict.hasOwnProperty(attachment_id)) {
            mapping_dict[getAttachmentId(context, sub_id, attachment_id, "get")]
              = attachment_id;
          }
        }
        for (attachment_id in result) {
          if (result.hasOwnProperty(attachment_id)) {
            if (mapping_dict.hasOwnProperty(attachment_id)) {
              attachments[mapping_dict[attachment_id]] = {};
            } else {
              attachments[attachment_id] = {};
            }
          }
        }
        return attachments;
      });
  };

  MappingStorage.prototype.hasCapacity = function (name) {
    return this._sub_storage.hasCapacity(name);
  };

  MappingStorage.prototype.repair = function () {
    return this._sub_storage.repair.apply(this._sub_storage, arguments);
  };

  MappingStorage.prototype.bulk = function (id_list) {
    var context = this;

    function mapId(parameter) {
      return getSubStorageId(context, parameter.parameter_list[0])
        .push(function (id) {
          return {"method": parameter.method, "parameter_list": [id]};
        });
    }

    return new RSVP.Queue()
      .push(function () {
        var promise_list = id_list.map(mapId);
        return RSVP.all(promise_list);
      })
      .push(function (id_list_mapped) {
        return context._sub_storage.bulk(id_list_mapped);
      })
      .push(function (result) {
        var mapped_result = [], i;
        for (i = 0; i < result.length; i += 1) {
          mapped_result.push(mapToMainDocument(
            context,
            result[i],
            false,
            true
          ));
        }
        return mapped_result;
      });
  };

  MappingStorage.prototype.buildQuery = function (option) {
    var context = this,
      i,
      query,
      property,
      select_list = [],
      sort_on = [];

    function mapQuery(one_query) {
      var j, mapped_kv, query_list = [];
      if (one_query.type === "complex") {
        for (j = 0; j < one_query.query_list.length; j += 1) {
          query_list.push(mapQuery(one_query.query_list[j]));
        }
        one_query.query_list = query_list;
        return one_query;
      }

      mapped_kv = mapProperty(context._mapping_dict[one_query.key],
                                  one_query.key, one_query.value, false);
      if (Array.isArray(mapped_kv.value) && mapped_kv.value.length) {
        query_list = [];
        for (j = 0; j < mapped_kv.value.length; j += 1) {
          query_list.push(new SimpleQuery({
            key: mapped_kv.key,
            value: mapped_kv.value[j],
            type: "simple"
          }));
        }
        one_query = new ComplexQuery({
          operator: "OR",
          query_list: query_list,
          type: "complex"
        });
        return one_query;
      }
      one_query.key = mapped_kv.key;
      one_query.value = mapped_kv.value;
      return one_query;
    }

    if (option.sort_on !== undefined) {
      for (i = 0; i < option.sort_on.length; i += 1) {
        property = mapProperty(this._mapping_dict[option.sort_on[i][0]],
                                  option.sort_on[i][0], undefined, false).key;
        if (property && sort_on.indexOf(property) < 0) {
          select_list.push([property, option.sort_on[i][1]]);
        }
      }
    }
    if (this._query.sort_on !== undefined) {
      for (i = 0; i < this._query.sort_on.length; i += 1) {
        property = mapProperty(this._mapping_dict[this._query.sort_on[i]],
                                  this._query.sort_on[i], undefined, false).key;
        if (sort_on.indexOf(property) < 0) {
          select_list.push([property, option.sort_on[i][1]]);
        }
      }
    }
    if (option.select_list !== undefined) {
      for (i = 0; i < option.select_list.length; i += 1) {
        property = mapProperty(this._mapping_dict[option.select_list[i]],
                                  option.select_list[i], undefined, false).key;
        if (property && select_list.indexOf(property) < 0) {
          select_list.push(property);
        }
      }
    }
    if (this._query.select_list !== undefined) {
      for (i = 0; i < this._query.select_list; i += 1) {
        property = this._query.select_list[i];
        if (select_list.indexOf(property) < 0) {
          select_list.push(property);
        }
      }
    }
    if (this._id_is_mapped) {
      select_list.push(this._mapping_dict.id.equal);
    }
    if (option.query !== undefined) {
      query = mapQuery(QueryFactory.create(option.query));
    }

    if (this._query.query !== undefined) {
      if (query === undefined) {
        query = this._query.query;
      }
      query = new ComplexQuery({
        operator: "AND",
        query_list: [query, this._query.query],
        type: "complex"
      });
    }

    if (query !== undefined) {
      query = Query.objectToSearchText(query);
    }
    return this._sub_storage.allDocs(
      {
        query: query,
        select_list: select_list,
        sort_on: sort_on,
        limit: option.limit
      }
    )
      .push(function (result) {
        for (i = 0; i < result.data.total_rows; i += 1) {
          result.data.rows[i].value =
            mapToMainDocument(
              context,
              result.data.rows[i].value,
              false,
              false
            );
          if (context._id_is_mapped) {
            result.data.rows[i].id =
                result.data.rows[i].value.id;
          }
          delete result.data.rows[i].value.id;
        }
        return result.data.rows;
      });
  };

  jIO.addStorage('mapping', MappingStorage);
}(jIO, RSVP, UriTemplate, SimpleQuery, ComplexQuery, QueryFactory, Query));